
open "partition_h"
open "rel_h"
open "gensym_h"

val partition_count = ref 0

exception InstClauseVar of string
exception InstClauseSort of string
exception InstClauseFsig of string
exception DoInstAtom of string

fun instantiate_clauses (f, atom_vars, fsig, invc) =
    fn (dcl as DCLAUSE (name, env, h, body)) =>
       let val vars = atom_vars h U union {atom_vars a
					  | a in list body}
	   (* the variables to replace by terms f(x1,...,xn)
	    for each possible constructor f of the corresponding sort. *)
	   fun iter_subst ({}, subst, rho, repl) =
	       f (definite_clause_subst subst (rho, dcl), repl)
	     | iter_subst ({x} U vars', subst, rho, repl) =
	       let val srt = ?env x handle MapGet => raise InstClauseVar x
	       in
		   iterate
		     let val (_, sl) = ?fsig c handle MapGet => raise InstClauseFsig c
			 val newvars = [V (gensym "_X") | _ in list sl]
			 val newenv = {x => s
				      || s in list sl and V x in list newvars}
		     in
			 iter_subst (vars', subst ++ {x => c $ newvars},
				     rho ++ newenv, c :: repl)
		     end
		   | c in set ?invc srt handle MapGet => raise InstClauseSort srt
		   end
	       end
       in
	   if empty vars
	       then f (dcl, nil)
	   else iter_subst (vars, {}, {}, nil)
       end

fun partition_clauses (pred_re, sort_re) =
    let val |[matches = predmatch, ...]| = regexp pred_re
	val |[matches = sortmatch, ...]| = regexp sort_re
    in
	fn (fsig, PROG (predsig, pr)) =>
	   let val constrs = {f => srt
			     | f => (srt, _) in map fsig
			       such that sortmatch srt}
	       val invc = invrel constrs
		   (* maps sort to be partitioned to non-empty set
		    of their constructors. *)
	       val psig = ref predsig
	       val pr' = ref {}
	       fun add dcl = pr' := add_gclause (!pr', dcl)
	       fun atom_vars (ATOM (P, l)) =
		   if predmatch P
		       then let val Psort = ?(!psig) P
				    handle MapGet => raise NoPredSig P
			    in
				{x
				|| V x in list l and srt in list Psort
				   such that sortmatch srt}
			    end
		   else {}
	       fun do_inst_atom (a as ATOM (P, l)) =
		   if predmatch P
		       then let val Psort = ?(!psig) P
				    handle MapGet => raise NoPredSig P
				val newP = ref false
				val f as |[put, convert, ...]| = outstring P
				val args =
				    (put ":";
				     print f (pack (!partition_count));
				     put "@";
				     append [if sortmatch srt
						 then case t of
							  f $ l' =>
							  (put "-";
							   put f;
							   newP := true;
							   l')
							| V x => raise DoInstAtom x
					     else [t]
					    || t in list l and srt in list Psort])
				val P' = convert ()
			    in
				if !newP
				    then let val argsorts =
						 append [if sortmatch srt
							     then case t of
								      f $ _ =>
								      (#2 (?fsig f)
								       handle MapGet =>
								       raise InstClauseFsig f)
								    | V x => raise DoInstAtom x
							 else [srt]
							|| t in list l and srt in list Psort]
					 in
					     psig := !psig ++ {P' => argsorts};
					     ATOM (P', args)
					 end
				else a
			    end
		   else a
	       fun do_inst (DCLAUSE (name, env, h, body), repl) =
		   let val f as |[put, convert, ...]| = outstring name
		   in
		       put ":";
		       print f (pack (!partition_count));
		       iterate
			 (put "-";
			  put c)
		       | c in list repl
		       end;
		       add (DCLAUSE (convert (), env, do_inst_atom h,
				     [do_inst_atom a | a in list body]))
		   end
	       val inst = instantiate_clauses (do_inst, atom_vars, fsig, invc)
	   in
	       inc partition_count;
	       iterate
		 iterate
		   inst dcl
		 | dcl in list dcls
		 end
	       | P => dcls in map pr
	       end;
	       PROG (!psig, !pr')
	   end
    end;

