
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "transition_h";
open "../Clauses/gensym_h";

    (* invert a transition relation (of type eva_program): *)
fun inv_trans {} = {}
  | inv_trans {x => ys} = {y => {(x, i)} | (y, i) in list ys}
  | inv_trans m = let val (m1, m2) = split m
		      val im1 = inv_trans m1
		      val im2 = inv_trans m2
		  in
		      (im1 delta im2) ++ {y => xs1 U xs2
					 | y => xs2 in map im1 <| im2
					     val xs1 = ?im1 y}
		  end;

exception PatDefinedVars of string
exception UnknownOp of string

fun pat_defined_vars env =
    let fun pat_defvars (s, T_VAR x) = {x => s}
	  | pat_defvars (_, T_APP ("exact", _)) = {}
	  | pat_defvars (_, T_APP ("crypt", [_, m, _])) =
	    pat_defvars (VAR_TYPE "number", m)
	  | pat_defvars (_, T_APP ("tuple", l)) =
	    overwrite [pat_defvars (VAR_TYPE "number", pat)
		      | pat in list l]
	  | pat_defvars (_, T_APP ("vanilla", nil)) = {}
	  | pat_defvars (_, T_APP ("d", [pat])) = pat_defvars (VAR_TYPE "*D*", pat)
	  | pat_defvars (_, T_APP ("p", [pat])) = pat_defvars (VAR_TYPE "principal", pat)
	  | pat_defvars (_, T_APP ("a", [pat])) = pat_defvars (VAR_TYPE "*algo*", pat)
	  | pat_defvars (_, T_APP ("sa", [pat])) = pat_defvars (VAR_TYPE "sym_algo", pat)
	  | pat_defvars (_, T_APP ("aa", [pat])) = pat_defvars (VAR_TYPE "asym_algo", pat)
	  | pat_defvars (_, T_APP (opp, args)) =
	    if opp inset {"apply", "secret-apply"}
		then
		    (let val T_VAR f = hd args
			 val l = tl args
			 val FUN_TYPE |[args, ...]| = ?env f
		     in
			 overwrite [pat_defvars (VAR_TYPE s, pat)
				   || pat in list l and s in list args]
		     end handle _ => raise PatDefinedVars opp)
	    else if opp inset {"apply-pubk", "apply-privk"}
		then (let val args' = tl args
			  val T_VAR f = hd args'
			  val l = tl args'
			  val FUN_TYPE |[args, ...]| = ?env f
		      in
			 overwrite [pat_defvars (VAR_TYPE s, pat)
				   || pat in list l and s in list args]
		      end handle _ => raise PatDefinedVars opp)
	    else raise UnknownOp opp
    in
	pat_defvars
    end;

(* same as pat_defined_vars, except does not need an environment, and
 does not return mappings from defined variables to their types, only
 sets of defined variables. *)
fun pat_defined_vars_set (T_VAR x) = {x}
  | pat_defined_vars_set (T_APP ("exact", _)) = {}
  | pat_defined_vars_set (T_APP ("crypt", [_, m, _])) =
    pat_defined_vars_set m
  | pat_defined_vars_set (T_APP ("tuple", l)) =
    union {pat_defined_vars_set pat | pat in list l}
  | pat_defined_vars_set (T_APP ("vanilla", nil)) = {}
  | pat_defined_vars_set (T_APP ("d", [pat])) = pat_defined_vars_set pat
  | pat_defined_vars_set (T_APP ("p", [pat])) = pat_defined_vars_set pat
  | pat_defined_vars_set (T_APP ("a", [pat])) = pat_defined_vars_set pat
  | pat_defined_vars_set (T_APP ("sa", [pat])) = pat_defined_vars_set pat
  | pat_defined_vars_set (T_APP ("aa", [pat])) = pat_defined_vars_set pat
  | pat_defined_vars_set (T_APP (opp, args)) =
    if opp inset {"apply", "secret-apply"}
	then
	    (let val l = tl args
	     in
		 union {pat_defined_vars_set pat | pat in list l}
	     end handle _ => raise PatDefinedVars opp)
    else if opp inset {"apply-pubk", "apply-privk"}
	then (let val l = tl (tl args)
	      in
		  union {pat_defined_vars_set pat | pat in list l}
	      end handle _ => raise PatDefinedVars opp)
    else raise UnknownOp opp;

(*exception UnknownAction of string*)

(* computes set of variables defined by action given in argument. *)
fun defined_vars pat_defvars =
    let fun defvars (T_APP ("new", [pat])) = pat_defvars (VAR_TYPE "*D*", pat)
	  | defvars (T_APP ("let", [pat, t])) =
	    pat_defvars (VAR_TYPE "number", pat)
	    (*!!! not sure that semantics of PEVA mandates that lets only bind numbers. *)
	  | defvars (T_APP ("recv", [pat])) =
	    pat_defvars (VAR_TYPE "number", pat)
	  | defvars _ = {}
    in
	defvars
    end;

fun defined_vars_with_types env = defined_vars (pat_defined_vars env);
val defined_vars_set = defined_vars (fn (_, pat) => pat_defined_vars_set pat);

exception PatUsedVars of string;

fun used_vars (T_APP ("send", [m])) = etvars m
  | used_vars (T_APP ("recv", [pat])) = pat_used_vars pat
  | used_vars (T_APP ("let", [pat, m])) = etvars m U pat_used_vars pat
  | used_vars _ = {}
and pat_used_vars (T_APP ("exact", [m])) = etvars m
  | pat_used_vars (T_APP ("crypt", [algo, _, k])) =
    etvars algo U etvars k
  | pat_used_vars (T_APP ("tuple", l)) = patl_used_vars l
  | pat_used_vars (T_APP ("d", [pat])) = pat_used_vars pat
  | pat_used_vars (T_APP ("p", [pat])) = pat_used_vars pat
  | pat_used_vars (T_APP ("a", [pat])) = pat_used_vars pat
  | pat_used_vars (T_APP ("sa", [pat])) = pat_used_vars pat
  | pat_used_vars (T_APP ("aa", [pat])) = pat_used_vars pat
  | pat_used_vars (T_APP ("vanilla", nil)) = {}
  | pat_used_vars (T_APP (opp, args)) =
    if opp inset {"apply", "secret-apply"}
	then
	    (let val f = hd args
		 val l = tl args
	     in
		 etvars f U patl_used_vars l
	     end handle _ => raise PatUsedVars opp)
    else if opp inset {"apply-pubk", "apply-privk"}
	then (let val algo = hd args
		  val args' = tl args
		  val f = hd args'
		  val l = tl args'
	      in
		  etvars algo U etvars f U patl_used_vars l
	      end handle _ => raise PatUsedVars opp)
    else raise UnknownOp opp
and patl_used_vars nil = {}
  | patl_used_vars (pat::l) =
    pat_used_vars pat U (patl_used_vars l \ pat_defined_vars_set pat)
    (*!!! complexity not optimal here: quadratic, could be linear. *)

exception TryInvertKey;

fun try_invert_key (algo, T_APP (opp, algo'::f::l)) =
    let val oppinv = ?{ "apply-pubk" => "apply-privk",
			"hash-apply-pubk" => "hash-apply-privk",
			"apply-privk" => "apply-pubk",
			"hash-apply-privk" => "hash-apply-pubk"
			} opp
    in
	(T_APP (oppinv, algo'::f::l),
	 if algo=algo'
	     then {}
	 else {(algo, algo')})
    end handle MapGet => raise TryInvertKey;

exception PatGenericTerm

    (* pat_generic_term rho pat renames all defined variables of pattern pat,
     building a generic term that matches the instances of pat.
     It also replaces subpatterns crypt(algo,pat,k) by a term crypt(algo,pat,k^{-1}),
     and possible side-conditions relating k and k^{-1} according to algo.
     *)
fun pat_generic_term rho =
    let val defined = ref {}
	val inverses = ref {}
	val equations = ref {}
	val rho = ref {}
	fun update m = etsubst (!defined) m
	fun patgt (T_VAR x) = if x inset !defined
				  then let val y = ?(!defined) x
				       in
					   defined := !defined ++ {x => y};
					   y
				       end
			      else T_VAR x
	  | patgt (T_APP ("exact", [m])) = update m
	  | patgt (T_APP ("crypt", [a, pat, k])) =
	    let val a' = update a
		val k' = update k
	    in
		case a of
		    T_APP ("a", [T_APP ("sa", _)]) => (* symmetric encryption: no key to invert. *)
		    T_APP ("crypt", [a', patgt pat, k'])
		  | T_APP ("a", [algo as T_APP ("aa", _)]) => (* asymmetric encryption: try to invert k. *)
		    (let val (invk, eqns) = try_invert_key (algo, k)
		     in
			 equations := !equations U eqns;
			 T_APP ("crypt", [a', patgt pat, invk])
		     end handle TryInvertKey => decrypt (a', patgt pat, k'))
		  | _ => decrypt (a', patgt pat, k')
	    end
	  | patgt (T_APP ("tuple", l)) =
	    T_APP ("tuple", [patgt pat | pat in list l])
	  | patgt (T_APP ("vanilla", nil)) = T_APP ("vanilla", nil)
	  | patgt (T_APP ("d", [pat])) = T_APP ("d", [patgt pat])
	  | patgt (T_APP ("p", [pat])) = T_APP ("p", [patgt pat])
	  | patgt (T_APP ("a", [pat])) = T_APP ("a", [patgt pat])
	  | patgt (T_APP ("sa", [pat])) = T_APP ("sa", [patgt pat])
	  | patgt (T_APP ("aa", [pat])) = T_APP ("aa", [patgt pat])
	  | patgt (T_APP (opp, args)) =
	    if opp inset {"apply", "secret-apply"}
		then (let val T_VAR f = hd args
			  val l = tl args
		      in
			  T_APP (opp, T_VAR f::[patgt pat | pat in list l])
		      end handle _ => raise PatGenericTerm)
	    else if opp inset {"apply-pubk", "apply-privk"}
		then (let val (algo :: T_VAR f :: l) = args
		      in
			  T_APP (opp, algo :: T_VAR f:: [patgt pat | pat in list l])
		      end handle _ => raise PatGenericTerm)
	    else raise UnknownOp opp
	and decrypt (a, m, k) =
	    let val newk = gensym "_invkey-"
		val k' = T_VAR newk
	    in
		inverses := !inverses U {(a, k, k')};
		rho := !rho ++ {newk => VAR_TYPE "number"};
		T_APP ("crypt", [a, m, k'])
	    end
    in
	fn pat => (defined := {}; inverses := {}; equations := {}; rho := {};
		   (patgt pat, !inverses, !equations, !rho))
    end;

fun live_vars (tr, observed) : state -m> string set =
    let val invtr = inv_trans tr
	val live = ref ({} : state -m> string set)
	(* for every instruction i,
	 gen(i) = {variables used by i}
	 kill(i) = {variables (re)defined by i}.
	 Then live (s) = union {(live (t) - kill (i)) U gen (i)
	                       | s --i--> t in tr}.
	 This is a least fixed point.
	 *)
	val worklist = ref (nil : (state * eva_term * state) list)
	memofun defined i = defined_vars_set i
	memofun used i = used_vars i
    in
	iterate
	  (live := !live ++ {s => observed};
	   iterate
	     (live := !live ++ {t => observed};
	      worklist := (s, i, t) :: !worklist)
	   | (t, i) in list succs
	   end)
	| s => succs in map tr
	end;
	while not (null (!worklist)) do
	    let val (s, i, t) = hd (!worklist)
		    before worklist := tl (!worklist)
		val lv = !live
		val oldlive_s = ?lv s
		val newlive_s = oldlive_s U (?lv t \ defined i) U used i
	    in
		live := lv ++ {s => newlive_s};
		if oldlive_s<>newlive_s andalso s inset invtr
		    then iterate
			worklist := (s', i', s) :: !worklist
			 | (s', i') in set ?invtr s
			 end
		else ()
	    end;
	    !live
    end;

fun declared_vars (name, initrho, init, tr) : state -m> type_env =
    let val declared = ref ({} : state -m> type_env)
	val worklist = ref [(init, initrho)]
	val defvars = defined_vars_with_types initrho
    in
	while not (null (!worklist)) do
	    let val (s, newrho) = hd (!worklist)
		    before worklist := tl (!worklist)
		val oldrho = if s inset !declared
				 then SOME (? (!declared) s)
			     else NONE
		val rho = case oldrho of
			      SOME rho => {x => typ
					  | x => typ in map (newrho <| rho)
					    such that ?newrho x = typ}
			    | _ => newrho
	    in
		declared := !declared ++ {s => rho};
		if oldrho<>SOME rho andalso s inset tr
		    then iterate
			worklist := (t, rho ++ defvars i) :: !worklist
			 | (t, i) in list ?tr s
			 end
		else ()
	    end;
	    !declared
    end;

fun new_sites tr =
    union {{(s, t)
	   | (t, T_APP ("new", _)) in list l}
	  | s => l in map tr};

fun stateP (role, s) =
    let val f as |[put, convert, ...]| = outstring "reach-"
    in
	put role;
	put "-";
	put s;
	convert ()
    end;

fun newF (role, s, t) =
    let val f as |[put, convert, ...]| = outstring "new-"
    in
	put role;
	put "-";
	put s;
	put "-";
	put t;
	convert ()
    end;
