
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "process_h";

extern val inv_trans : eva_program -> string -m> (string * eva_term) set;
    (* invert a transition relation (of type eva_program): *)

extern val pat_defined_vars : type_env -> eva_type * eva_term -> type_env;
	   (* given environment (used to give signatures to functions in *apply* patterns),
	    type T and pattern P of type T, compute set of variables defined by P,
	    together with their types. *)
extern val pat_defined_vars_set : eva_term -> string set;
	   (* given pattern P, compute set of defined variables (no types). *)

extern val defined_vars_with_types : type_env -> eva_term -> type_env;
    (* as pat_defined_vars, except second argument is an action
     (labeling transitions). *)
extern val defined_vars_set : eva_term -> string set;
    (* as pat_defined_vars_set, except second argument is an action. *)

extern val used_vars : eva_term -> string set;
    (* given action A, return set of variables it uses. *)
extern val pat_used_vars : eva_term -> string set;
(* given pattern P, return set of variables it uses. *)

extern val live_vars : eva_program * string set -> state -m> string set;
	   (* given program P, and set V of variables, computes
	    map that associates each state S with the set of variables live at S.
	    It is considered that all variables in V are live everywhere. *)
extern val declared_vars : string * type_env * state * eva_program -> state -m> type_env;
	   (* given principal name A, initial typing environment E, initial state S0, and program P,
	    computes map that associates each state S
	    with set of variables that have been defined on every path leading
	    to S, together with their types. *)

extern val new_sites : eva_program -> (state * state) set;
	   (* computes sets of pairs (s, t) such that s-->t is a NEW transition
	    in given transition relation. *)

extern val stateP : string * state -> string;
	   (* given role name R and state S, produce name for a
	    reachability predicate reach-<R>-<S>. *)
extern val newF : string * state * state -> string;
	   (* given role name R, states S and T, produce name
	    for a Skolem function that represents the nonces
	    created by role R during transition S --> T.
	    This function is new-<R>-<S>-<T>. *)

extern val pat_generic_term : eva_subst -> eva_term
	   -> eva_term * (eva_term * eva_term * eva_term) set * (eva_term * eva_term) set * type_env;
	   (* Given a substitution rho and a pattern P, such that
	    rho maps the defined variables of P to terms,
	    returns a 4-tuple (T, inverses, equations, rho), where:
	    - T is a generic term matching pat,
	    - inverses is a set of triples (a, k, k') such that k and k' should be inverses
	    for algorithm a [a being of type number, not *algo*],
	    - equations is a set of pairs (a, a') of encryption algorithms that should
	    match [of type *algo* here],
	    - rho is a typing environment described newly introduced variables for expressing
	    inverses. *)
