
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

(* sorting functions *)

fun merge less =
    let fun merge ([],l2) = l2
	  | merge (l1,[]) = l1
	  | merge (l1 as x1::l'1,l2 as x2::l'2) =
	    if less(x1,x2)
		then x1::merge(l'1,l2)
	    else x2::merge(l1,l'2)
    in
	merge
    end

fun sort less = (* (''a * ''a -> bool) -> (''a -m> 'b) -> ''a list *)
    let val merge = merge less
	fun sort {} = []
	  | sort {x => _} = [x]
	  | sort vars =
	    let val (vars1,vars2) = split vars
	    in
		merge (sort vars1,sort vars2)
	    end
    in
	sort
    end

val string_sort = sort (op strless);

fun least less =
    let fun find {} = raise Empty
	  | find {x => _} = x
	  | find xs = let val (xs1,xs2) = split xs
			  val x1 = find xs1
			  val x2 = find xs2
		      in
			  if less (x1,x2)
			      then x1
			  else x2
		      end
    in
	find
    end
