
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "term_h"
open "evaterm_h"
open "conclo_h"

extern val trans_warn : string * lineinfo -> unit;
extern val trans_fatal : string * lineinfo -> 'a;

datatype eva_type = VAR_TYPE of string
  | FUN_TYPE of |[ args : string list,
		   res : string,
		   hash : bool,
		   secret : bool
		   ]|;

extern val print_eva_type : 'a outstream -> eva_type -> unit;

type type_env = string -m> eva_type;
type value_env = string -m> eva_term;

extern val fvars : eva_term -> string set;
extern val eva_subst : value_env -> eva_term -> eva_term;

exception FailEvt of unit cont option * string;

extern val trans_type : term -> eva_type;

type trans_term_loc_share_info = (*;;;*)
    (string * string * eva_term -> eva_term) (* to make "located" terms from "locate" terms *)
    * (string -> unit) (* to check "shared" variables. *);

extern val trans_term : trans_term_loc_share_info option * type_env * value_env
	   -> term -> eva_term * eva_type;
	   (* trans_term (do_locate, env, sigma) t
	    translates the (Lisp) term t with variables typed by env,
	    possibly aliased as in sigma, returns translation to eva term
	    with corresponding most general type. *)

extern val trans_term_type : trans_term_loc_share_info option * type_env * value_env
	   -> term * eva_type -> eva_term;
       (* trans_term_type (do_locate, env, sigma) (t, typ)
	same as trans_term, except coerces t to have type typ.
	*)

type coercions = eva_type -m> eva_type -m> eva_term -> eva_term;

extern val complete_coercions : coercions -> coercions;
extern val eva_coercions : unit -> coercions;

extern val msplit : term -> term * term;
	   (* tente de recuperer les parties emetteur et recepteur
	    d'un message intervenant dans le protocole. *)

type eva_knowledge = eva_term -m> eva_term;

extern val trans_compose : type_env * (eva_term -> eva_term) * (string -> exn) * 'a cc ->
	   eva_knowledge * string set * eva_term -> eva_knowledge * eva_term list;
extern val trans_expect : type_env * 'a cc -> eva_knowledge * eva_term ->
	   eva_knowledge * eva_term list;

type eva_program = string -m> (string * eva_term) set;

extern val eva_prog_subst : value_env -> eva_program -> eva_program;

type axiom = eva_term * eva_term * eva_type * type_env;

datatype eva_decl = EVA_DECL of |[ env : type_env,
				   shared : string set, (*;;;*)
				   sigma : value_env,
				   local_sigma : string -m> value_env,
				   everybody : eva_knowledge,
				   knows : string -m> eva_knowledge,
				   axioms : axiom set,
				   equations : eva_type cc
				   ]|;

datatype id_set = ID_INCL of string set
  | ID_EXCL of string set;

extern val trans_decl : term -> eva_decl;
extern val trans_code : type_env * value_env * (string -m> value_env) *
	   eva_knowledge * (string -m> eva_knowledge) * axiom set * eva_type cc ->
	   term -> string * (string -m> eva_program * eva_knowledge);
extern val trans_sessions : type_env * value_env -> term -> (string -m> string set option * value_env * id_set);

type process = string set option (* privates *)
    * eva_program
    * eva_knowledge;

extern val trans_properties : type_env * string set * value_env * (string -m> process)
	   -> term list -> eva_term list;

datatype eva_spec = EVA_SPEC of |[ env : type_env,
				   sigma : value_env,
				   axioms : (eva_term * eva_term * eva_type * type_env) set,
				   system : string * (string -m> process),
				   assume : eva_term list,
				   claim : eva_term list
				   ]|;

extern val translate : term -> eva_spec;
extern val convert_eva_spec : eva_spec -> term;
