{

(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open Lisptoken

let read_octal s i =
	let n = String.length s in
	let ir = ref i in
	let acc = ref 0 in
	(while !ir<n do
	   acc := !acc lsl 3 + Char.code (String.get s (!ir)) - 48;
	   ir := !ir+1
         done;
         !acc);;

let string_buf = Buffer.create 256
}

let space = [' ' '\t' '\n' '\r']+
let dig = ['0'-'9']+
let id = [^ '"' ' ' '\t' '\n' '\r' '(' ')']+
rule lisptoken = parse
    space { lisptoken lexbuf }
  | '('   { LISP_LPAREN }
  | ')'   { LISP_RPAREN }
  | dig   { LISP_INT (int_of_string (Lexing.lexeme lexbuf)) }
  | id    { LISP_ID (Lexing.lexeme lexbuf) }
  | "\""
      { Buffer.reset string_buf;
        string lexbuf;
	LISP_STRING (Buffer.contents string_buf)
      }
  | eof { raise LispEof }
  | _ { raise (LispInvalidChar (Lexing.lexeme lexbuf)) }
and string = parse
    '"' { () }
  | '\n'+ { string lexbuf }
  | '\\' ['0'-'7'] ['0'-'7']? ['0'-'7']? { Buffer.add_char string_buf (Char.chr (read_octal (Lexing.lexeme lexbuf) 1)) }
  | '\\' ['0'-'7']+ { raise (LispBadEscape (Lexing.lexeme lexbuf)) }
  | '\\' 'n' { Buffer.add_char string_buf '\n'; string lexbuf }
  | '\\' 't' { Buffer.add_char string_buf '\t'; string lexbuf }
  | '\\' 'r' { Buffer.add_char string_buf '\r'; string lexbuf }
  | '\\' 'b' { Buffer.add_char string_buf '\b'; string lexbuf }
  | '\\' 'f' { Buffer.add_char string_buf '\014'; string lexbuf }
  | '\\' _ { Buffer.add_char string_buf (Lexing.lexeme_char lexbuf 1); string lexbuf }
  | [^ '\\' '\n' '"']+ { Buffer.add_string string_buf (Lexing.lexeme lexbuf); string lexbuf }
  | _ { Buffer.add_char string_buf (Lexing.lexeme_char lexbuf 0); string lexbuf }
  | eof { raise LispEofInString }
